# Orca
#
# Copyright 2020 The Orca Team
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Object for maintaining the state of the object navigator."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2020 The Orca Team"
__license__   = "LGPL"

import pyatspi

# compatibility layer
orca_state = None
def _(text):
    # for compatibility til we have translation
    return text


class ObjectNavigator:
    def __init__(self, API):
        self.navigatorFocus=None
        self.lastNavigatorFocus=None
        self.lastLocusOfFocus=None
        self.simplify=True
        self.inclusionList = [pyatspi.ROLE_PARAGRAPH]
        self.exclusionList = [pyatspi.ROLE_SECTION]
        self.API = API
        global orca_state
        orca_state = self.API.app.getDynamicApiManager().getAPI('OrcaState')

    def _isEmpty(self, object):
        """ Method for determining whether an object should be excluded from simple navigation. """
        if object.role in self.inclusionList:
            return False
        return orca_state.activeScript.utilities.isLayoutOnly(object) or object.getRole() in orca_state.activeScript.utilities.getCellRoles() or object.role in self.exclusionList

    def _children(self, object):
        """ Returns a list of children for object, taking simple navigation into account if enabled. """
        if object.childCount==0: # if we have no children, return the empty list
            return []
        if not self.simplify: # just return the usual children
            return list(object)
        children = list(object) # the normal list of children
        resultingChildren = [] # the list we're constructing
        for child in children:
            if self._isEmpty(child): # add the children of empty objects to our list of children
                resultingChildren.extend(self._children(child))
            else:
                resultingChildren.append(child)
        return resultingChildren

    def _parent(self, object):
        """ Returns the parent for object, taking simple navigation into account if enabled. """
        if not self.simplify: # return the usual parent
            return object.parent
        while object.parent is not None and self._isEmpty(object.parent): # move up until we find an object with a suitible parent
            object = object.parent
        return object.parent

    def setNavigatorFocus(self, object):
        """ Method for changing the navigator focus. Placing this in a method allow us to keep track of the previous focus. """
        self.lastNavigatorFocus=self.navigatorFocus
        self.navigatorFocus=object

    def update(self):
        """ Updates the navigator focus to the last keyboard focus, if it is different from before. """
        if orca_state.locusOfFocus==self.lastLocusOfFocus:
            return  
        self.navigatorFocus=orca_state.locusOfFocus
        self.lastLocusOfFocus=orca_state.locusOfFocus

    def present(self):
        """ Presents the current navigator focus to the user. """
        generatorMap={}
        if self.lastNavigatorFocus:
            generatorMap['priorObj']=self.lastNavigatorFocus
        utterances = orca_state.activeScript.speechGenerator.generateSpeech(self.navigatorFocus, **generatorMap)
        if len(utterances)>0:
            orca_state.activeScript.presentItemsInSpeech(utterances)
        else:
            orca_state.activeScript.speakMessage(self.navigatorFocus.getLocalizedRoleName())
        orca_state.activeScript.displayBrailleForObject(self.navigatorFocus)

    def up(self):
        """ Moves the navigator focus to the parent of the current focus. """
        self.update()   
        parent = self._parent(self.navigatorFocus)
        if parent is not None:
            self.setNavigatorFocus(parent)
            self.present()
        else:
            orca_state.activeScript.speakMessage(_("No parent."))

    def down(self):
        """ Moves the navigator focus to the first child of the current focus. """
        self.update()
        children = self._children(self.navigatorFocus)
        if len(children)>0:
            self.setNavigatorFocus(children[0])
            self.present()
        else:
            orca_state.activeScript.speakMessage(_("No children."))

    def next(self):
        """ Moves the navigator focus to the next sibling of the current focus. """
        self.update()
        parent = self._parent(self.navigatorFocus)
        if parent==None:
            orca_state.activeScript.speakMessage(_("No next."))
            return
        siblings = self._children(parent)
        if self.navigatorFocus in siblings:
            index = siblings.index(self.navigatorFocus)
            if index<len(siblings)-1:
                self.setNavigatorFocus(siblings[index+1])
                self.present()
            else:
                orca_state.activeScript.speakMessage(_("No next."))
        else:
            self.setNavigatorFocus(parent)

    def previous(self):
        """ Moves the navigator focus to the previous sibling of the current focus. """
        self.update()   
        parent = self._parent(self.navigatorFocus)
        if parent==None:
            orca_state.activeScript.speakMessage(_("No previous."))
            return
        siblings = self._children(parent)
        if self.navigatorFocus in siblings:
            index = siblings.index(self.navigatorFocus)
            if index>0:
                self.setNavigatorFocus(siblings[index-1])
                self.present()
            else:
                orca_state.activeScript.speakMessage(_("No previous."))
        else:
            self.setNavigatorFocus(parent) 
